pca <- function(object, k=2, scale=c("none", "pareto", "vector", "UV"),
                center=TRUE, limit=1e-6, maxiterations=5000, varLimit=1, subset=numeric(),
                method=c("svd", "nipals"), verbose=FALSE,...) {
  
  if(inherits(object, "exprSet")) 
    object <- t(exprs(object))

  scale <- match.arg(scale)
  method <- match.arg(method)

  missing <- sum(is.na(object))

  if(length(subset) > 0)
    object <- object[,subset]

  if(center) {
    object <- scale(object, scale=FALSE, center=TRUE)
  }
  if(scale != "none") {
    switch(scale,
           UV = {
             object <- scale(object, scale=TRUE, center=FALSE)
           },
           vector = {
             vectorNorm <- function(y) {
               y <- y / sqrt(sum(y^2))
               return(y)
             }
             object <- t(apply(object, 1, vectorNorm))
           },
           pareto = {
             object <- apply(object, 2,
                             function(object) {
                               return(object / sqrt(sd(object, na.rm=TRUE)))
                             })
           })
  }

  if(missing > 0 & method == "svd") {
    warning("Found missing values, using the nipals method instead of requested svd")
    method <- "nipals"
  }

  switch(method,
         svd = {
           pcs <- prcomp(object, center=FALSE, scale.=FALSE)
           imp <- summary(pcs)$importance
           if(varLimit < 1)
             k <- sum(imp[3,] < varLimit) + 1
           r <- new("pcaRes", scores=cbind(pcs$x[,1:k]), loadings=cbind(pcs$rotation[,1:k]),
                    R2cum=imp[3,1:k], sdev=pcs$sdev, nobs=nrow(object), nvar=ncol(object),
                    R2=imp[2,1:k], centered=center, subset=subset, varLimit=varLimit,
                    scaled=scale, k=k, method="svd", missing=missing, center=colMeans(object))
           return(r)
         },
         nipals = {
           nobs <- nrow(object)
           nvar <- ncol(object)


           ##Find a good starting column
           ##(one where not all values are NA) and do some initial checking
           for (i in 1:nvar) {         
             if (length(na.omit(object[,i])) != 0) {
               startingColumn <- i
             }
             else {
               warning("Whole column was found NA in original matrix")
             }
           }
           for (i in 1:nobs) {
             if (length(na.omit(object[i,])) == 0) {
               warning("Whole row was found NA in original matrix")
             }
           }

           scaledobject <- object

           ph <- rep(0, nvar)
           R2cum <- NULL
           scores <- NULL
           loadings <- NULL
           anotherPc <- TRUE
           l <- 1
           
           while(anotherPc) {
             count <- 0                 #number of iterations done
             th <- object[,startingColumn]   #first column is starting vector for th
             continue <- TRUE
             if(verbose) cat(paste("Calculating PC", l, ": ", sep=""))
             
             while(continue) {
               count <- count+1
               ph <- rep(0, nvar)

               ##Calculate loadings through LS regression
               for (i in 1:nvar) {
                 ph[i] <- sum(object[,i] * th, na.rm=TRUE) / sum(th^2, na.rm=TRUE)
                 ##if the whole column is NA the loading of that variable should also be NA
                 if (length(na.omit(object[,i])) == 0) { 
                   ph[i] <- NA
                 }
               }
               ##normalize ph based on the available values.
               ph <- ph / sqrt(sum(na.omit(ph)^2))

               ##Calculate scores through LS regression
               th.old <- th
               th <- rep(0, nobs)
               for (i in 1:nobs) {
                 th[i] <- sum(object[i,] * ph, na.rm=TRUE) / sum(ph^2, na.rm=TRUE)
                 if (length(na.omit(object[i,])) == 0) {
                   th[i] <- NA
                 }
               }
               
               ##Round up by calculating if convergence condition is met and
               ##checking if it seems to be an neverending loop.
               if (count > maxiterations) {
                 stop("Too many iterations, quitting")
               }
               if (t(na.omit(th.old - th)) %*% (na.omit(th.old - th)) <= limit) {
                 continue = FALSE
               }
               if (verbose)cat("*")
             }
             if (verbose) cat(" Done\n")
             object <- object - (th %*% t(ph))
             scores <- cbind(scores, th)
             loadings <- cbind(loadings, ph)
             
             ##cumulative proportion of variance
             R2cum <- cbind(R2cum, 1 - (sum(object^2,na.rm=TRUE) / sum(scaledobject^2, na.rm=TRUE)))
             l <- l + 1
             if (R2cum[1,l - 1] >= varLimit || l > k) {
               anotherPc <- FALSE
               k <- l - 1
             }
           }
           R2 <- vector(length=length(R2cum), mode="numeric")
           R2[1] <- R2cum[1]
           if (length(R2cum) > 1) {
             for (i in 2:length(R2)) {
               R2[i] <- R2cum[i] - R2cum[i-1]
             }
           }
           rownames(scores) <- rownames(object)
           colnames(scores) <- paste("PC", 1:k, sep="")
           rownames(loadings) <- colnames(object)
           colnames(loadings) <- paste("PC", 1:k, sep="")
           r <- new("pcaRes", scores=scores, loadings=loadings, k=k,
                    R2cum=c(R2cum), nobs=nobs, nvar=nvar, R2=R2, sdev=apply(scores, 2, sd),
                    scaled=scale, centered=center, varLimit=varLimit, method="nipals",
                    subset=subset, missing=missing, center=colMeans(object))
           return(r)
         })
}

plotPcs <- function(object, pc=1:object@k, scoresLoadings=c(TRUE, FALSE),...) {
  if(length(dev.list()) == 0) 
    x11(width=16, height=10)
  ba <- (length(pc)*(length(pc) - 1)) / 2
  mf <- c(floor(sqrt(ba)),ceiling(sqrt(ba)))
  if(mf[1] * mf[2] < ba)
    mf[1] <- mf[1] + 1
  par(mfrow=mf,cex=0.9, mar=c(4,3,3,1))
  

  perm <- function(object) {
    res <- NULL
    tri <- matrix(TRUE, length(object), length(object))
    tri[upper.tri(tri, TRUE)] <- FALSE
    for(i in 1:length(object))
      for(j in 1:length(object))
        if(tri[i,j])
          res <- rbind(res, c(object[i],object[j]))
    res
  }

  pp <- perm(pc)

  for(i in 1:nrow(pp))
    slplot(object, pcs=pp[i,], scoresLoadings=scoresLoadings, rug=FALSE,...)
}

setMethod("print", "pcaRes",
          function(x, ...) {
            summary(x)
            cat(x@nvar, "\tVariables\n")
            cat(x@nobs,"\tSamples\n")
            cat(x@missing, "\tNA's\n")
            cat(x@k, "\tCalculated component(s)\n")
            cat("*** Data was ")
            if(x@scaled != "none") { cat("scaled (using ",x@scaled,") ", sep="") }
            else { cat("NOT scaled ") }
            if(x@centered && (x@scaled == "none")) { cat("but centered ***\n") }
            else if(x@centered && (x@scaled != "none")) { cat("and centered ***\n") }
            else if(!x@centered && (x@scaled != "none")) { cat("but NOT centered ***\n") }
            else if(!x@centered && !x@scaled) { cat("and NOT centered ***\n") }
            cat("Scores structure:\n")
            print(dim(x@scores))
            cat("Loadings structure:\n")
            print(dim(x@loadings))
          })

setMethod("summary", "pcaRes",
          function(object) {
            if (length(object@R2cum) != length(na.omit(object@R2cum))) {
              warning("Whole row or column was NA in original matriobject. Variance cannot be estimated")
            }
            cat(object@method, "calculated PCA\n")
            cat("Importance of component(s):\n")
            prop <- vector(length=length(object@R2cum), mode="numeric")
            prop[1] <- object@R2cum[1]
            if (length(object@R2cum) > 1) {
              for (i in 2:length(prop)) {
                prop[i] <- object@R2cum[i] - object@R2cum[i-1]
              }
            }
            r <- rbind(prop, object@R2cum)
            rownames(r) <- c("Proportion of Variance", "Cumulative Proportion")
            colnames(r) <- paste("PC", 1:object@k, sep="")
            print(r, digits=4)
            invisible(r)
          })

setMethod("screeplot", "pcaRes",
          function(x, npcs=min(10, length(x@R2)), type = c("barplot", "lines"),
                   main = deparse(substitute(x)), ...) {
            newx <- list(sdev=x@sdev)
            screeplot(newx, npcs, type, main,...)
          })

setMethod("slplot", "pcaRes",
          function(object, pcs=c(1,2), scoresLoadings=c(TRUE, TRUE), sl=rownames(object@scores),
                   ll=rownames(object@loadings), hotelling=0.95, rug=TRUE,...) {
            if(length(pcs) > 2)
              plotPcs(object, pcs, scoresLoadings=scoresLoadings,...)

            else {
              if(sum(scoresLoadings) == 2) 
                layout(matrix(c(1,2), 1, 2, TRUE), respect=matrix(c(1,1), 1, 2))
              if (length(pcs) == 1 || object@k == 1) {
                pcs <- 1
                if(scoresLoadings[1]) {
                  barplot(object@scores[,pcs], main="Scores",
                          sub=paste(sprintf("%.2f", object@R2cum[pcs] * 100),
                            "% of the variance explained", sep=""), las=3, names.arg=sl, ...)
                }
                if(scoresLoadings[2])
                  barplot(object@loadings[,pcs], main="Loadings", las=3, ...)
                return(TRUE)
              }
              if(scoresLoadings[1]) {
                if (!is.null(sl)) {
                  plot(object@scores[,pcs], type="n", main="Scores",
                       sub=paste(sprintf("%.2f",object@R2cum[max(pcs)] * 100),
                         "% of the variance explained", sep=""), ...)
                  text(object@scores[,pcs], sl,...)
                  if(rug)
                    rug(object@scores[,1])
                }
                else {
                  plot(object@scores[,pcs], main="Scores",
                       sub=paste(sprintf("%.2f",object@R2cum[max(pcs)] * 100),
                         "% of the variance explained", sep=""),...)
                }
                abline(h=0, v=0)
                if(!is.null(hotelling)) {
                  A <- length(pcs)
                  lines(ellipse(cov(object@scores[,pcs]), level=hotelling))
                }
              }
              if(scoresLoadings[2]) {
                if (!is.null(ll)) {
                  plot(object@loadings[,pcs], type="n", main="Loadings", ...)
                  text(object@loadings[,pcs], ll,...)
                }
                else {
                  plot(object@loadings[,pcs], main="Loadings", ...)
                }
                abline(h=0, v=0)
              }
            }
          })

